@@ -30,7 +30,8 @@ import numpy as np
 
 from typing import TYPE_CHECKING
 if TYPE_CHECKING:
-    from typing import Callable, Tuple, Union, Dict, List, Optional, Sequence, Generator
+    from typing import Callable, Tuple, Union, Dict, List, Optional, \
+        Sequence, Generator, Any, IO, Set
     Curve = Callable[[float], float]
 
 
@@ -46,13 +47,52 @@ def _isPulseaudioRunning() -> bool:
     if _re.search(r'\bConnection\ failure', output) is None:
         return True
     return False
-    #retcode = _subprocess.call(["pulseaudio", "--check"])
-    #return retcode == 0
 
 
 _audioDeviceRegex = r"(\d+):\s((?:adc|dac)\d+)\s*\((.*)\)(?:\s+\[ch:(\d+)\])?"
 
 
+def midiDevices(backend='portmidi') -> Tuple[List[MidiDevice], List[MidiDevice]]:
+    """
+    Returns input and output midi devices for the given backend
+
+    Args:
+        backend: the backend used for realtime midi (as passed to
+            csound via -+rtmidi={backend}
+
+    Returns:
+        a tuple (inputdevices, outputdevices), which each of these
+        is a list of MidiDevice with attributes ``deviceid`` (the value
+        passed to -M), ``name`` (the name of the device) and ``kind``
+        (one of 'input' or 'output')
+
+    ========   ===========================
+    Platform   Possible Backends
+    ========   ===========================
+    linux      portmidi, alsaseq, alsaraw
+    macos      portmidi
+    windows    portmidi
+    ========   ===========================
+    """
+
+    import ctcsound
+    csound = ctcsound.Csound()
+    csound.setOption(f"-+rtmidi={backend}")
+    csound.setOption("-odac")
+    csound.start()
+    inputdevs = csound.midiDevList(False)
+    outputdevs = csound.midiDevList(True)
+    logger.debug(f"MIDI Inputs:  {inputdevs}")
+    logger.debug(f"MIDI Outputs: {outputdevs}")
+    midiins = [MidiDevice(deviceid=d['device_id'], kind='input',
+                          name=f"{d['interface_name']}:{d['device_name']}")
+               for d in inputdevs]
+    midiouts = [MidiDevice(deviceid=d['device_id'], kind='output',
+                           name=f"{d['interface_name']}:{d['device_name']}")
+               for d in outputdevs]
+    return midiins, midiouts
+
+
 @dataclass(unsafe_hash=True)
 class AudioBackend:
     """
@@ -92,19 +132,11 @@ class AudioBackend:
         # we get the devices via getAudioDevices to enable caching
         indevs, outdevs = getAudioDevices(self.name)
         devs = indevs if kind == 'input' else outdevs
-        if pattern.startswith('dac:') or pattern.startswith('adc:'):
-            pattern = pattern[4:]
-            namesearch = True
-        elif _re.match('(dac|adc)[0-9]+', pattern):
-            namesearch = False
-        else:
-            namesearch = True
-        if not namesearch:
-            return next((d for d in devs
-                         if d.id == pattern), None)
-        else:
-            return next((d for d in devs
-                         if _re.search(pattern, d.name)), None)
+        match = next((d for d in devs if d.id == pattern), None)
+        if match:
+            return match
+        return next((d for d in devs
+                    if _re.search(pattern, d.name)), None)
 
     def isAvailable(self) -> bool:
         """ Is this backend available? """
@@ -134,6 +166,7 @@ class AudioBackend:
         """
         return (self.defaultBufferSize, self.defaultNumBuffers)
 
+
     def audioDevices(self) -> Tuple[List[AudioDevice], List[AudioDevice]]:
         """
         Returns a tuple (input devices, output devices)
@@ -155,8 +188,8 @@ class AudioBackend:
                 idxstr, devid, devname, numchannels = groups
                 numchannels = int(numchannels) if numchannels is not None else 2
             kind = 'input' if devid.startswith("adc") else 'output'
-            dev = AudioDevice(index=int(idxstr), id=devid, name=devname, kind=kind,
-                              numchannels=numchannels)
+            dev = AudioDevice(index=int(idxstr), id=devid.strip(), name=devname,
+                              kind=kind, numchannels=numchannels)
             (indevices if kind == 'input' else outdevices).append(dev)
         return indevices, outdevices
 
@@ -761,7 +794,7 @@ class TableDataFile:
         size: the size of the table
     """
     tabnum: int
-    data: Union[List[float], np.ndarray, str]
+    data: Sequence[float]
     """the data itself or a path to a file"""
     fmt: str = 'gen23'  # One of 'wav', 'gen23'
     start: float = 0
@@ -787,7 +820,8 @@ class TableDataFile:
             saveAsGen23(self.data, outfile=outfile)
         elif self.fmt in ('wav', 'aif', 'aiff', 'flac'):
             import sndfileio
-            sndfileio.sndwrite(outfile, self.data, sr=44100,
+            dataarr = np.asarray(self.data, dtype=float)
+            sndfileio.sndwrite(outfile, dataarr, sr=44100,
                                metadata={'comment': 'Datafile'})
 
     def scoreLine(self, outfile: str) -> str:
@@ -1094,6 +1128,20 @@ def saveMatrixAsGen23(outfile: str, mtx: np.ndarray, extradata:List[float]=None,
             f.write(rowstr)
             f.write("\n")
 
+@dataclass
+class MidiDevice:
+    """
+    A MidiDevice holds information about a midi device for a given backend
+
+    Attributes:
+        index: the index as listed by csound and passed to -M
+        name: the name of the device
+        kind: the kind of the device ('input', 'output')
+    """
+    deviceid: str
+    name: str
+    kind: str = 'input'
+
 
 @dataclass
 class AudioDevice:
@@ -1489,6 +1537,8 @@ class Csd:
     def strset(self, s:str) -> int:
         """
         Add a strset to this csd
+
+        If ``s`` has already been passed, the same index is returned
         """
         idx = self._str2index.get(s)
         if idx is not None:
@@ -2161,7 +2211,10 @@ def getNchnls(backend: str=None, outpattern:str=None, inpattern:str=None,
     else:
         outdev = backendDef.searchAudioDevice(outpattern, kind='output')
         if not outdev:
-            raise ValueError(f"Output device {outpattern} not found")
+            indevs, outdevs = backendDef.audioDevices()
+            outdevids = [d.id for d in outdevs]
+            raise ValueError(f"Output device '{outpattern}' not found. Possible devices "
+                             f"are: {outdevids}")
     nchnls = outdev.numchannels if outdev.numchannels is not None else defaultout
     if inpattern is None:
         indev = adc
